/*
 *  This file is part of athena-ssl-cipher-check.
 *
 *  athena-ssl-cipher-check is free software: you can redistribute it
 *  and/or modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation, either version 3 of the License,
 *  or (at your option) any later version.
 *
 *  athena-ssl-cipher-check is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
 *  Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with athena-ssl-cipher-check. If not, see
 *  <http://www.gnu.org/licenses/>.
 *
 */

package athenasslciphercheck;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.logging.Level;
import java.util.logging.Logger;

public class tlsHeuristics {

    private String host;
    private int port;

    private ArrayList<Cipher> cipherDescList;

    public int resetCount = 0;
    public int timedOutCount = 0;
    public int unidentifiedCount = 0;
    public int rejectedCount = 0;

    public void checkTlsCiphers()
    {
        //build cipher byte array
        //    possible ciphers * cipher code size = array size
        //    256^2 * 2 = 131072
        byte[] cipherList = new byte[131072];

        //build cipher list
        int count = 0;
        for( int x = 0; x <= 0xFF; x++ )
        {
            for( int y = 0; y <= 0xFF; y++ )
            {
                cipherList[count] = (byte) x;
                count++;
                cipherList[count] = (byte) y;
                count++;
            }
        }

        //start searching
        System.out.println();
        System.out.println("SSLv3 Ciphers");
        for( int i = 0x00; i <= (0xFF/32); i++ )
        {
            this.search(cipherList, 512*i*32, 512*32, (byte) 0x00);
            if( timedOutCount > 5 )
            {
                System.out.println("SSlv3 being skipped due to timeouts");
                break;
            }            
        }

        this.printStats();

        resetCount = 0;
        timedOutCount = 0;
        unidentifiedCount = 0;
        rejectedCount = 0;

        System.out.println();
        System.out.println("TLSv1 Ciphers");
        for( int i = 0x00; i <= (0xFF/32); i++ )
        {
            this.search(cipherList, 512*i*32, 512*32, (byte) 0x01);
            if( timedOutCount > 5 )
            {
                System.out.println("TLSv1 being skipped due to timeouts");
                break;
            }            
        }

        this.printStats();
    }

    public void search( byte[] cipherList, int offset, int length, byte ver )
    {

        int lowerResult = tlsCipherCheck.check(host, port, cipherList, offset, length/2, ver);
        int upperResult = tlsCipherCheck.check(host, port, cipherList, offset+(length/2), length/2, ver);

        updateCounts( lowerResult, upperResult );

        //if lower half of tree contains 1 cipher and was accepted
        if( length == 4 && lowerResult == 1 )
        {
            byte[] tmp = new byte[2];
            tmp[0] = cipherList[ offset ];
            tmp[1] = cipherList[ offset+1 ];

            Cipher c = this.lookUpCipher(tmp);
            if( c != null)
                System.out.println(c.getDesc());
            else
                System.out.println("UNKNOWN CIPHER # " + (offset/2) );
        }

        //if upper half of tree contains 1 cipher and was accepted
        if ( length == 4 && upperResult == 1 )
        {
            //System.out.println("Found one!! "+( (offset+(length/2)) / 2) );
            byte[] tmp = new byte[2];
            tmp[0] = cipherList[ offset+(length/2) ];
            tmp[1] = cipherList[ offset+(length/2) +1 ];

            Cipher c = this.lookUpCipher(tmp);
            if( c != null)
                System.out.println(c.getDesc());
            else
                System.out.println("UNKNOWN CIPHER #" + (offset+(length/4)) );

        }

        //If not at the end of the tree, continue searching on positive branches
        if( length != 4 )
        {
            if( lowerResult == 1 )
            {
                this.search(cipherList, offset, length/2, ver );
            }
            if( upperResult == 1 )
            {
                this.search(cipherList, offset+(length/2), length/2, ver );
            }
        }

    }

    public tlsHeuristics( String h, int p )
    {
        port = p;
        host = h;

        cipherDescList = new ArrayList<Cipher>();

        loadCiphers();
    }

    public void loadCiphers()
    {
        //ffs, There has got be a better way of doing this...
        String path = "";
        if( System.getProperty("java.class.path").contains( System.getProperty("file.separator") ) )
                path += System.getProperty("java.class.path").substring(0, 1+System.getProperty("java.class.path").lastIndexOf(System.getProperty("file.separator")));
        //</ffs>
        try {
            BufferedReader fileIn = new BufferedReader(new FileReader(path+"tlsciphers.txt"));
            String line;

            int count = 0; //line number counter
            while( (line = fileIn.readLine()) != null )
            {
                count++; //inc line number
                if( line.charAt(0) == '#' ) //if comment, ignore
                    continue;

                if( line.split(" ").length != 4 )
                {
                    System.err.println("Error in configuration file on line "+count+", 4 words expected.");
                    continue;
                }

                String cipherCode = line.split(" ")[0];
                if( cipherCode.length() != 4 ||
                    !this.isHexidec( cipherCode ) )
                {
                    System.err.println("Error in configuration file on line "+count
                                       +" Codes must be hexidecimal");
                    continue;
                }

                Cipher tmpCipher = new Cipher(  line.split(" ")[1],
                                                hexStringToByte( cipherCode ),
                                                Integer.parseInt( line.split(" ")[2] ),
                                                Integer.parseInt( line.split(" ")[3] ) );
                cipherDescList.add(tmpCipher);                
            }
            
            fileIn.close();
        }
        catch ( Exception ex) {
            System.out.println( ex.getMessage() );
            System.out.println("configuration file '"+path+"tlsciphers.txt' not found");
        }


    }

    private Cipher lookUpCipher( byte[] tmp )
    {
        Cipher rtn;
        for( int i = 0; i < cipherDescList.size(); i++ )
        {
            if( cipherDescList.get(i).getId()[0] == tmp[0] &&
                cipherDescList.get(i).getId()[1] == tmp[1] )
                return cipherDescList.get(i);
        }
        return null;
    }

    public boolean isHexidec( String s )
    {
        for( int i = 0; i < s.length(); i ++ )
        {
            char c = s.charAt(i);
            if( (c >= '0' && c <= '9') )
                continue;
            if( (c >= 'a' && c <= 'f') )
                continue;
            if( (c >= 'A' && c <= 'F') )
                continue;

            return false;
        }
        return true;
    }

    /*
     * Number of hexidecimal digits %2 must equal 0
     * i.e. an even number of characters
     */
    public byte[] hexStringToByte( String s )
    {
        byte[] array = new byte[s.length()/2];

        for( int i = 0; i < s.length(); i += 2 )
        {
            int tmp = 0;
            tmp = (hexValue( s.charAt(i) )*16) + hexValue( s.charAt(i+1) );
            array[i/2] = (byte) tmp;
        }

        return array;
    }

    public int hexValue(char c)
    {
        switch( c )
        {
            case '0':
                return 0;
            case '1':
                return 1;
            case '2':
                return 2;
            case '3':
                return 3;
            case '4':
                return 4;
            case '5':
                return 5;
            case '6':
                return 6;
            case '7':
                return 7;
            case '8':
                return 8;
            case '9':
                return 9;

            case 'A':
            case 'a':
                return 10;

            case 'B':
            case 'b':
                return 11;

            case 'C':
            case 'c':
                return 12;

            case 'D':
            case 'd':
                return 13;

            case 'E':
            case 'e':
                return 14;

            case 'F':
            case 'f':
                return 15;

            default:
                return -1;
        }
    }

    private void printStats()
    {
        if( /*resetCount == 0 &&*/
            timedOutCount == 0 &&
            unidentifiedCount == 0 /*&&
            rejectedCount == 0*/ )
            return;

        System.out.println("");
        boolean first = true;
        if( false /*resetCount > 0*/ )
        {
            System.out.print(resetCount + " connections reset");
            first = false;
        }

        if( timedOutCount > 0 )
        {
            if( !first )
                System.out.print(", ");
            System.out.print(timedOutCount + " connections timed out");
            first = false;
        }

        if( unidentifiedCount > 0 )
        {
            if( !first )
                System.out.print(", ");
            System.out.print(unidentifiedCount + " connections responded with unexpected data");
            first = false;
        }

        if( /*rejectedCount > 0*/ false )
        {
            if( !first )
                System.out.print(", ");
            System.out.print(rejectedCount + " ciphers sets rejected");
            first = false;
        }

        System.out.println();
    }

    private void updateCounts( int lowerResult, int upperResult)
    {
        if( lowerResult == 0 || lowerResult == -5 )
            resetCount++;
        if( upperResult == 0 || upperResult == -5  )
            resetCount++;

        if( lowerResult == -1 )
            rejectedCount++;
        if( upperResult == -1 )
            rejectedCount++;

        if( lowerResult == -2 )
            unidentifiedCount++;
        if( upperResult == -2 )
            unidentifiedCount++;

        if( lowerResult == -3 )
        {
            timedOutCount++;
        }
        if( upperResult == -3 )
        {
            timedOutCount++;
        }
    }
}
